import io
import mimetypes
import uuid
from collections import abc
from http.cookiejar import CookieJar
from json import dumps as json_dumps
from typing import (
    Any,
    Callable,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Tuple,
    Union,
)
from urllib.parse import quote_plus
from urllib.parse import urlencode as _urlencode

from hyperlink import DecodedURL, EncodedURL
from requests.cookies import merge_cookies
from treq.cookies import scoped_cookie
from twisted.internet.defer import Deferred
from twisted.internet.interfaces import IProtocol
from twisted.python.components import proxyForInterface, registerAdapter
from twisted.python.filepath import FilePath
from twisted.web.client import (
    BrowserLikeRedirectAgent,
    ContentDecoderAgent,
    CookieAgent,
    FileBodyProducer,
    GzipDecoder,
    IAgent,
    RedirectAgent,
)
from twisted.web.http_headers import Headers
from twisted.web.iweb import IBodyProducer, IResponse

from treq import multipart
from treq._types import (
    _CookiesType,
    _DataType,
    _FilesType,
    _FileValue,
    _HeadersType,
    _ITreqReactor,
    _JSONType,
    _ParamsType,
    _URLType,
)
from treq.auth import add_auth
from treq.response import _Response


class _Nothing:
    """Type of the sentinel `_NOTHING`"""


_NOTHING = _Nothing()


def urlencode(query: _ParamsType, doseq: bool) -> bytes:
    s = _urlencode(query, doseq)
    return s.encode("ascii")


def _scoped_cookiejar_from_dict(
    url_object: EncodedURL, cookie_dict: Optional[Mapping[str, str]]
) -> CookieJar:
    """
    Create a CookieJar from a dictionary whose cookies are all scoped to the
    given URL's origin.

    @note: This does not scope the cookies to any particular path, only the
        host, port, and scheme of the given URL.
    """
    cookie_jar = CookieJar()
    if cookie_dict is None:
        return cookie_jar
    for k, v in cookie_dict.items():
        cookie_jar.set_cookie(scoped_cookie(url_object, k, v))
    return cookie_jar


class _BodyBufferingProtocol(proxyForInterface(IProtocol)):  # type: ignore
    def __init__(self, original, buffer, finished):
        self.original = original
        self.buffer = buffer
        self.finished = finished

    def dataReceived(self, data: bytes) -> None:
        self.buffer.append(data)
        self.original.dataReceived(data)

    def connectionLost(self, reason: Exception) -> None:
        self.original.connectionLost(reason)
        self.finished.errback(reason)


class _BufferedResponse(proxyForInterface(IResponse)):  # type: ignore
    def __init__(self, original):
        self.original = original
        self._buffer = []
        self._waiters = []
        self._waiting = None
        self._finished = False
        self._reason = None

    def _deliverWaiting(self, reason):
        self._reason = reason
        self._finished = True
        for waiter in self._waiters:
            for segment in self._buffer:
                waiter.dataReceived(segment)
            waiter.connectionLost(reason)

    def deliverBody(self, protocol):
        if self._waiting is None and not self._finished:
            self._waiting = Deferred()
            self._waiting.addBoth(self._deliverWaiting)
            self.original.deliverBody(
                _BodyBufferingProtocol(protocol, self._buffer, self._waiting)
            )
        elif self._finished:
            for segment in self._buffer:
                protocol.dataReceived(segment)
            protocol.connectionLost(self._reason)
        else:
            self._waiters.append(protocol)


class HTTPClient:
    def __init__(
        self,
        agent: IAgent,
        cookiejar: Optional[CookieJar] = None,
        data_to_body_producer: Callable[[Any], IBodyProducer] = IBodyProducer,
    ) -> None:
        self._agent = agent
        if cookiejar is None:
            cookiejar = CookieJar()
        self._cookiejar = cookiejar
        self._data_to_body_producer = data_to_body_producer

    def get(self, url: _URLType, **kwargs: Any) -> "Deferred[_Response]":
        """
        See :func:`treq.get()`.
        """
        kwargs.setdefault("_stacklevel", 3)
        return self.request("GET", url, **kwargs)

    def put(
        self, url: _URLType, data: Optional[_DataType] = None, **kwargs: Any
    ) -> "Deferred[_Response]":
        """
        See :func:`treq.put()`.
        """
        kwargs.setdefault("_stacklevel", 3)
        return self.request("PUT", url, data=data, **kwargs)

    def patch(
        self, url: _URLType, data: Optional[_DataType] = None, **kwargs: Any
    ) -> "Deferred[_Response]":
        """
        See :func:`treq.patch()`.
        """
        kwargs.setdefault("_stacklevel", 3)
        return self.request("PATCH", url, data=data, **kwargs)

    def post(
        self, url: _URLType, data: Optional[_DataType] = None, **kwargs: Any
    ) -> "Deferred[_Response]":
        """
        See :func:`treq.post()`.
        """
        kwargs.setdefault("_stacklevel", 3)
        return self.request("POST", url, data=data, **kwargs)

    def head(self, url: _URLType, **kwargs: Any) -> "Deferred[_Response]":
        """
        See :func:`treq.head()`.
        """
        kwargs.setdefault("_stacklevel", 3)
        return self.request("HEAD", url, **kwargs)

    def delete(self, url: _URLType, **kwargs: Any) -> "Deferred[_Response]":
        """
        See :func:`treq.delete()`.
        """
        kwargs.setdefault("_stacklevel", 3)
        return self.request("DELETE", url, **kwargs)

    def request(
        self,
        method: str,
        url: _URLType,
        *,
        params: Optional[_ParamsType] = None,
        headers: Optional[_HeadersType] = None,
        data: Optional[_DataType] = None,
        files: Optional[_FilesType] = None,
        json: Union[_JSONType, _Nothing] = _NOTHING,
        auth: Optional[Tuple[Union[str, bytes], Union[str, bytes]]] = None,
        cookies: Optional[_CookiesType] = None,
        allow_redirects: bool = True,
        browser_like_redirects: bool = False,
        unbuffered: bool = False,
        reactor: Optional[_ITreqReactor] = None,
        timeout: Optional[float] = None,
        _stacklevel: int = 2,
    ) -> "Deferred[_Response]":
        """
        See :func:`treq.request()`.
        """
        method_: bytes = method.encode("ascii").upper()

        if isinstance(url, DecodedURL):
            parsed_url = url.encoded_url
        elif isinstance(url, EncodedURL):
            parsed_url = url
        elif isinstance(url, str):
            # We use hyperlink in lazy mode so that users can pass arbitrary
            # bytes in the path and querystring.
            parsed_url = EncodedURL.from_text(url)
        else:
            parsed_url = EncodedURL.from_text(url.decode("ascii"))

        # Join parameters provided in the URL
        # and the ones passed as argument.
        if params:
            parsed_url = parsed_url.replace(
                query=parsed_url.query + tuple(_coerced_query_params(params))
            )

        url = parsed_url.to_uri().to_text().encode("ascii")

        headers = self._request_headers(headers, _stacklevel + 1)

        bodyProducer, contentType = self._request_body(
            data, files, json, stacklevel=_stacklevel + 1
        )
        if contentType is not None:
            headers.setRawHeaders(b"Content-Type", [contentType])

        if not isinstance(cookies, CookieJar):
            cookies = _scoped_cookiejar_from_dict(parsed_url, cookies)

        merge_cookies(self._cookiejar, cookies)
        wrapped_agent: IAgent = CookieAgent(self._agent, self._cookiejar)

        if allow_redirects:
            if browser_like_redirects:
                wrapped_agent = BrowserLikeRedirectAgent(wrapped_agent)
            else:
                wrapped_agent = RedirectAgent(wrapped_agent)

        wrapped_agent = ContentDecoderAgent(wrapped_agent, [(b"gzip", GzipDecoder)])

        if auth:
            wrapped_agent = add_auth(wrapped_agent, auth)

        d = wrapped_agent.request(
            method_, url, headers=headers, bodyProducer=bodyProducer
        )

        if reactor is None:
            from twisted.internet import reactor  # type: ignore
        assert reactor is not None

        if timeout:
            delayedCall = reactor.callLater(timeout, d.cancel)

            def gotResult(result):
                if delayedCall.active():
                    delayedCall.cancel()
                return result

            d.addBoth(gotResult)

        if not unbuffered:
            d.addCallback(_BufferedResponse)

        return d.addCallback(_Response, self._cookiejar)

    def _request_headers(
        self, headers: Optional[_HeadersType], stacklevel: int
    ) -> Headers:
        """
        Convert the *headers* argument to a :class:`Headers` instance
        """
        if isinstance(headers, dict):
            h = Headers({})
            for k, v in headers.items():
                if isinstance(v, (bytes, str)):
                    h.addRawHeader(k, v)
                elif isinstance(v, list):
                    h.setRawHeaders(k, v)
                else:
                    raise TypeError(
                        "The value of headers key {!r} has non-string type {}.".format(
                            k, type(v)
                        )
                    )
            return h
        if isinstance(headers, Headers):
            return headers
        if headers is None:
            return Headers({})

        raise TypeError(
            (
                "headers must be a dict, twisted.web.http_headers.Headers, or None,"
                " but found {}."
            ).format(type(headers))
        )

    def _request_body(
        self,
        data: Optional[_DataType],
        files: Optional[_FilesType],
        json: Union[_JSONType, _Nothing],
        stacklevel: int,
    ) -> Tuple[Optional[IBodyProducer], Optional[bytes]]:
        """
        Here we choose a right producer based on the parameters passed in.

        :params data:
            Arbitrary request body data.

            If *files* is also passed this must be a :class:`dict`,
            a :class:`tuple` or :class:`list` of field tuples as accepted by
            :class:`MultiPartProducer`. The request is assigned a Content-Type
            of ``multipart/form-data``.

            If a :class:`dict`, :class:`list`, or :class:`tuple` it is
            URL-encoded and the request assigned a Content-Type of
            ``application/x-www-form-urlencoded``.

            Otherwise, any non-``None`` value is passed to the client's
            *data_to_body_producer* callable (by default,
            :class:`IBodyProducer`), which accepts file-like objects.

        :params files:
            Files to include in the request body, in any of the several formats
            described in :func:`_convert_files()`.

        :params json:
            JSON-encodable data, or the sentinel `_NOTHING`. The sentinel is
            necessary because ``None`` is a valid JSON value.
        """
        if json is not _NOTHING:
            if files or data:
                raise TypeError(
                    "Argument 'json' cannot be combined with '{}'.".format(
                        "data" if data else "files"
                    )
                )
            return (
                self._data_to_body_producer(
                    json_dumps(json, separators=(",", ":")).encode("utf-8"),
                ),
                b"application/json; charset=UTF-8",
            )

        if files:
            # If the files keyword is present we will issue a
            # multipart/form-data request as it suits better for cases
            # with files and/or large objects.
            fields: List[Tuple[str, _FileValue]] = []
            if data:
                for field in _convert_params(data):
                    fields.append(field)
            for field in _convert_files(files):
                fields.append(field)

            boundary = str(uuid.uuid4()).encode("ascii")
            return (
                multipart.MultiPartProducer(fields, boundary=boundary),
                b"multipart/form-data; boundary=" + boundary,
            )

        # Otherwise stick to x-www-form-urlencoded format
        # as it's generally faster for smaller requests.
        if isinstance(data, (dict, list, tuple)):
            return (
                # FIXME: The use of doseq here is not permitted in the types, and
                # sequence values aren't supported in the files codepath. It is
                # maintained here for backwards compatibility. See
                # https://github.com/twisted/treq/issues/360.
                self._data_to_body_producer(urlencode(data, doseq=True)),
                b"application/x-www-form-urlencoded",
            )
        elif data:
            return (
                self._data_to_body_producer(data),
                None,
            )

        return None, None


def _convert_params(params: _DataType) -> Iterable[Tuple[str, str]]:
    items_method = getattr(params, "items", None)
    if items_method:
        return list(sorted(items_method()))
    elif isinstance(params, (tuple, list)):
        return list(params)
    else:
        raise ValueError("Unsupported format")


def _convert_files(files):
    """
    Files can be passed in a variety of formats:

    * {"fieldname": open("bla.f", "rb")}
    * {"fieldname": ("filename", open("bla.f", "rb"))}
    * {"fieldname": ("filename", "content-type", open("bla.f", "rb"))}
    * Anything that has iteritems method, e.g. MultiDict:
      MultiDict([(name, open()), (name, open())]

    Our goal is to standardize it to unified form of:

    * [(param, (file name, content type, producer))]
    """

    if hasattr(files, "iteritems"):
        files = files.iteritems()
    elif hasattr(files, "items"):
        files = files.items()

    for param, val in files:
        file_name, content_type, fobj = (None, None, None)
        if isinstance(val, tuple):
            if len(val) == 2:
                file_name, fobj = val
            elif len(val) == 3:
                file_name, content_type, fobj = val
            else:
                # NB: This is TypeError for backward compatibility. This case
                # used to fall through to `IBodyProducer`, below, which raised
                # TypeError about being unable to coerce None.
                raise TypeError(
                    (
                        "`files` argument must be a sequence of tuples of"
                        " (file_name, file_obj) or"
                        " (file_name, content_type, file_obj),"
                        " but the {!r} tuple has length {}: {!r}"
                    ).format(param, len(val), val),
                )
        else:
            fobj = val
            if hasattr(fobj, "name"):
                file_name = FilePath(fobj.name).basename()

        if not content_type:
            content_type = _guess_content_type(file_name)

        # XXX: Shouldn't this call self._data_to_body_producer?
        yield (param, (file_name, content_type, IBodyProducer(fobj)))


def _query_quote(v: Any) -> str:
    """
    Percent-encode a querystring name or value.

    :param v: A value.

    :returns:
        The value, coerced to a string and percent-encoded as appropriate for
        a querystring (with space as ``+``).
    """
    if not isinstance(v, (str, bytes)):
        v = str(v)
    if not isinstance(v, bytes):
        v = v.encode("utf-8")
    q = quote_plus(v)
    return q


def _coerced_query_params(params: _ParamsType) -> Iterator[Tuple[str, str]]:
    """
    Carefully coerce *params* in the same way as `urllib.parse.urlencode()`

    Parameter names and values are coerced to unicode, which is encoded as
    UTF-8 and then percent-encoded. As a special case, `bytes` are directly
    percent-encoded.

    :param params:
        A mapping or sequence of (name, value) two-tuples. The value may be
        a list or tuple of multiple values. Names and values may be pretty much
        any type.

    :returns:
        A generator that yields two-tuples containing percent-encoded text
        strings.
    """
    items: Iterable[Tuple[str, Union[str, Tuple[str, ...], List[str]]]]
    if isinstance(params, abc.Mapping):
        items = params.items()
    else:
        items = params

    for key, values in items:
        key_quoted = _query_quote(key)

        if not isinstance(values, (list, tuple)):
            values = (values,)
        for value in values:
            yield key_quoted, _query_quote(value)


def _from_bytes(orig_bytes: bytes) -> IBodyProducer:
    return FileBodyProducer(io.BytesIO(orig_bytes))


def _from_file(orig_file: Union[io.BytesIO, io.BufferedReader]) -> IBodyProducer:
    return FileBodyProducer(orig_file)


def _guess_content_type(filename: str) -> Optional[str]:
    if filename:
        guessed = mimetypes.guess_type(filename)[0]
    else:
        guessed = None
    return guessed or "application/octet-stream"


registerAdapter(_from_bytes, bytes, IBodyProducer)
registerAdapter(_from_file, io.BytesIO, IBodyProducer)

# file()/open() equiv
registerAdapter(_from_file, io.BufferedReader, IBodyProducer)



"""
Provides treq version information.
"""

# This file is auto-generated! Do not edit!
# Use `incremental` to change this file.

from incremental import Version

__version__ = Version("treq", 24, 9, 1)
__all__ = ["__version__"]



import unittest
from unittest import mock
from typing import Optional

from twisted.python.failure import Failure

from twisted.internet.error import ConnectionDone
from twisted.trial.unittest import TestCase
from twisted.web.http_headers import Headers
from twisted.web.client import ResponseDone, ResponseFailed
from twisted.web.http import PotentialDataLoss
from twisted.web.resource import Resource
from twisted.web.server import NOT_DONE_YET

from treq import collect, content, json_content, text_content
from treq.content import _encoding_from_headers
from treq.client import _BufferedResponse
from treq.testing import StubTreq


class ContentTests(TestCase):
    def setUp(self):
        self.response = mock.Mock()
        self.protocol = None

        def deliverBody(protocol):
            self.protocol = protocol

        self.response.deliverBody.side_effect = deliverBody
        self.response = _BufferedResponse(self.response)

    def test_collect(self):
        data = []

        d = collect(self.response, data.append)

        self.protocol.dataReceived(b'{')
        self.protocol.dataReceived(b'"msg": "hell')
        self.protocol.dataReceived(b'o"}')

        self.protocol.connectionLost(Failure(ResponseDone()))

        self.assertEqual(self.successResultOf(d), None)

        self.assertEqual(data, [b'{', b'"msg": "hell', b'o"}'])

    def test_collect_failure(self):
        data = []

        d = collect(self.response, data.append)

        self.protocol.dataReceived(b'foo')

        self.protocol.connectionLost(Failure(ResponseFailed("test failure")))

        self.failureResultOf(d, ResponseFailed)

        self.assertEqual(data, [b'foo'])

    def test_collect_failure_potential_data_loss(self):
        """
        PotentialDataLoss failures are treated as success.
        """
        data = []

        d = collect(self.response, data.append)

        self.protocol.dataReceived(b'foo')

        self.protocol.connectionLost(Failure(PotentialDataLoss()))

        self.assertEqual(self.successResultOf(d), None)

        self.assertEqual(data, [b'foo'])

    def test_collect_0_length(self):
        self.response.length = 0

        d = collect(
            self.response,
            lambda d: self.fail("Unexpectedly called with: {0}".format(d)))

        self.assertEqual(self.successResultOf(d), None)

    def test_content(self):
        d = content(self.response)

        self.protocol.dataReceived(b'foo')
        self.protocol.dataReceived(b'bar')
        self.protocol.connectionLost(Failure(ResponseDone()))

        self.assertEqual(self.successResultOf(d), b'foobar')

    def test_content_cached(self):
        d1 = content(self.response)

        self.protocol.dataReceived(b'foo')
        self.protocol.dataReceived(b'bar')
        self.protocol.connectionLost(Failure(ResponseDone()))

        self.assertEqual(self.successResultOf(d1), b'foobar')

        def _fail_deliverBody(protocol):
            self.fail("deliverBody unexpectedly called.")

        self.response.original.deliverBody.side_effect = _fail_deliverBody

        d3 = content(self.response)

        self.assertEqual(self.successResultOf(d3), b'foobar')

        self.assertNotIdentical(d1, d3)

    def test_content_multiple_waiters(self):
        d1 = content(self.response)
        d2 = content(self.response)

        self.protocol.dataReceived(b'foo')
        self.protocol.connectionLost(Failure(ResponseDone()))

        self.assertEqual(self.successResultOf(d1), b'foo')
        self.assertEqual(self.successResultOf(d2), b'foo')

        self.assertNotIdentical(d1, d2)

    def test_json_content(self):
        self.response.headers = Headers()
        d = json_content(self.response)

        self.protocol.dataReceived(b'{"msg":"hello!"}')
        self.protocol.connectionLost(Failure(ResponseDone()))

        self.assertEqual(self.successResultOf(d), {"msg": "hello!"})

    def test_json_content_unicode(self):
        """
        When Unicode JSON content is received, the JSON text should be
        correctly decoded.
        RFC7159 (8.1): "JSON text SHALL be encoded in UTF-8, UTF-16, or UTF-32.
        The default encoding is UTF-8"
        """
        self.response.headers = Headers()
        d = json_content(self.response)

        self.protocol.dataReceived(u'{"msg":"hëlló!"}'.encode('utf-8'))
        self.protocol.connectionLost(Failure(ResponseDone()))

        self.assertEqual(self.successResultOf(d), {u'msg': u'hëlló!'})

    def test_json_content_utf16(self):
        """
        JSON received is decoded according to the charset given in the
        Content-Type header.
        """
        self.response.headers = Headers({
            b'Content-Type': [b"application/json; charset='UTF-16LE'"],
        })
        d = json_content(self.response)

        self.protocol.dataReceived(u'{"msg":"hëlló!"}'.encode('UTF-16LE'))
        self.protocol.connectionLost(Failure(ResponseDone()))

        self.assertEqual(self.successResultOf(d), {u'msg': u'hëlló!'})

    def test_text_content(self):
        self.response.headers = Headers(
            {b'Content-Type': [b'text/plain; charset=utf-8']})

        d = text_content(self.response)

        self.protocol.dataReceived(b'\xe2\x98\x83')
        self.protocol.connectionLost(Failure(ResponseDone()))

        self.assertEqual(self.successResultOf(d), u'\u2603')

    def test_text_content_default_encoding_no_param(self):
        self.response.headers = Headers(
            {b'Content-Type': [b'text/plain']})

        d = text_content(self.response)

        self.protocol.dataReceived(b'\xa1')
        self.protocol.connectionLost(Failure(ResponseDone()))

        self.assertEqual(self.successResultOf(d), u'\xa1')

    def test_text_content_default_encoding_no_header(self):
        self.response.headers = Headers()

        d = text_content(self.response)

        self.protocol.dataReceived(b'\xa1')
        self.protocol.connectionLost(Failure(ResponseDone()))

        self.assertEqual(self.successResultOf(d), u'\xa1')

    def test_content_application_json_default_encoding(self):
        self.response.headers = Headers(
            {b'Content-Type': [b'application/json']})

        d = text_content(self.response)

        self.protocol.dataReceived(b'gr\xc3\xbcn')
        self.protocol.connectionLost(Failure(ResponseDone()))

        self.assertEqual(self.successResultOf(d), u'grün')

    def test_text_content_unicode_headers(self):
        """
        Parsing of the Content-Type header is robust in the face of
        Unicode gunk in the header value, which is ignored.
        """
        self.response.headers = Headers(
            {
                b"Content-Type": [
                    'text/plain; charset="UTF-16BE"; u=ᛃ'.encode("utf-8")
                ],
            }
        )

        d = text_content(self.response)

        self.protocol.dataReceived(u'ᚠᚡ'.encode('UTF-16BE'))
        self.protocol.connectionLost(Failure(ResponseDone()))

        self.assertEqual(self.successResultOf(d), u'ᚠᚡ')


class UnfinishedResponse(Resource):
    """Write some data, but never finish."""

    isLeaf = True

    def __init__(self):
        Resource.__init__(self)
        # Track how requests finished.
        self.request_finishes = []

    def render(self, request):
        request.write(b"HELLO")
        request.notifyFinish().addBoth(self.request_finishes.append)
        return NOT_DONE_YET


class MoreRealisticContentTests(TestCase):
    """Tests involving less mocking."""

    def test_exception_handling(self):
        """
        An exception in the collector function:

            1. Always gets returned in the result ``Deferred`` from
               ``treq.collect()``.

            2. Closes the transport.
        """
        resource = UnfinishedResponse()
        stub = StubTreq(resource)
        response = self.successResultOf(stub.request("GET", "http://127.0.0.1/"))
        self.assertEqual(response.code, 200)

        def error(data):
            1 / 0

        d = collect(response, error)

        # Exceptions in the collector are passed on to the caller via the
        # response Deferred:
        self.failureResultOf(d, ZeroDivisionError)

        # An exception in the protocol results in the transport for the request
        # being closed.
        stub.flush()
        self.assertEqual(len(resource.request_finishes), 1)
        self.assertIsInstance(resource.request_finishes[0].value, ConnectionDone)


class EncodingFromHeadersTests(unittest.TestCase):
    def _encodingFromContentType(self, content_type: str) -> Optional[str]:
        """
        Invoke `_encoding_from_headers()` for a header value.

        :param content_type: A Content-Type header value.
        :returns: The result of `_encoding_from_headers()`
        """
        h = Headers({"Content-Type": [content_type]})
        return _encoding_from_headers(h)

    def test_rfcExamples(self):
        """
        The examples from RFC 9110 § 8.3.1 are normalized to
        canonical (lowercase) form.
        """
        for example in [
            "text/html;charset=utf-8",
            'Text/HTML;Charset="utf-8"',
            'text/html; charset="utf-8"',
            "text/html;charset=UTF-8",
        ]:
            self.assertEqual("utf-8", self._encodingFromContentType(example))

    def test_multipleParams(self):
        """The charset parameter is extracted even if mixed with other params."""
        for example in [
            "a/b;c=d;charSet=ascii",
            "a/b;c=d;charset=ascii; e=f",
            "a/b;c=d; charsEt=ascii;e=f",
            "a/b;c=d;   charset=ascii;  e=f",
        ]:
            self.assertEqual("ascii", self._encodingFromContentType(example))

    def test_quotedString(self):
        """Any quotes that surround the value of the charset param are removed."""
        self.assertEqual(
            "ascii", self._encodingFromContentType("foo/bar; charset='ASCII'")
        )
        self.assertEqual(
            "shift_jis", self._encodingFromContentType('a/b; charset="Shift_JIS"')
        )

    def test_noCharset(self):
        """None is returned when no valid charset parameter is found."""
        for example in [
            "application/octet-stream",
            "text/plain;charset=",
            "text/plain;charset=''",
            "text/plain;charset=\"'\"",
            "text/plain;charset=🙃",
        ]:
            self.assertIsNone(self._encodingFromContentType(example))



# Copyright (c) Twisted Matrix Laboratories.
# See LICENSE for details.

from typing import cast, AnyStr

from io import BytesIO

from .._multipart import MultipartParser
from twisted.trial import unittest
from zope.interface.verify import verifyObject

from twisted.internet import task
from twisted.internet.testing import StringTransport
from twisted.web.client import FileBodyProducer
from twisted.web.iweb import UNKNOWN_LENGTH, IBodyProducer

from treq.multipart import MultiPartProducer, _LengthConsumer


class MultiPartProducerTestCase(unittest.TestCase):
    """
    Tests for the L{MultiPartProducer} which gets dictionary like object
    with post parameters, converts them to multipart/form-data format
    and feeds them to an L{IConsumer}.
    """
    def _termination(self):
        """
        This method can be used as the C{terminationPredicateFactory} for a
        L{Cooperator}.  It returns a predicate which immediately returns
        C{False}, indicating that no more work should be done this iteration.
        This has the result of only allowing one iteration of a cooperative
        task to be run per L{Cooperator} iteration.
        """
        return lambda: True

    def setUp(self):
        """
        Create a L{Cooperator} hooked up to an easily controlled, deterministic
        scheduler to use with L{MultiPartProducer}.
        """
        self._scheduled = []
        self.cooperator = task.Cooperator(
            self._termination, self._scheduled.append)

    def getOutput(self, producer, with_producer=False):
        """
        A convenience function to consume and return output.
        """
        consumer = output = BytesIO()

        producer.startProducing(consumer)

        while self._scheduled:
            self._scheduled.pop(0)()

        if with_producer:
            return (output.getvalue(), producer)
        else:
            return output.getvalue()

    def newLines(self, value: AnyStr) -> AnyStr:

        if isinstance(value, str):
            return value.replace(u"\n", u"\r\n")
        else:
            return value.replace(b"\n", b"\r\n")

    def test_interface(self):
        """
        L{MultiPartProducer} instances provide L{IBodyProducer}.
        """
        self.assertTrue(
            verifyObject(
                IBodyProducer, MultiPartProducer({})))

    def test_unknownLength(self) -> None:
        """
        If the L{MultiPartProducer} is constructed with a file-like object
        passed as a parameter without either a C{seek} or C{tell} method,
        its C{length} attribute is set to C{UNKNOWN_LENGTH}.
        """
        class CantTell:
            def seek(self, offset, whence):
                """
                A C{seek} method that is never called because there is no
                matching C{tell} method.
                """

        class CantSeek:
            def tell(self):
                """
                A C{tell} method that is never called because there is no
                matching C{seek} method.
                """

        producer = MultiPartProducer(
            {"f": ("name", "application/octet-stream", FileBodyProducer(CantTell()))})
        self.assertEqual(UNKNOWN_LENGTH, producer.length)

        producer = MultiPartProducer(
            {"f": ("name", "application/octet-stream", FileBodyProducer(CantSeek()))})
        self.assertEqual(UNKNOWN_LENGTH, producer.length)

    def test_knownLengthOnFile(self) -> None:
        """
        If the L{MultiPartProducer} is constructed with a file-like object with
        both C{seek} and C{tell} methods, its C{length} attribute is set to the
        size of the file as determined by those methods.
        """
        inputBytes = b"here are some bytes"
        inputFile = BytesIO(inputBytes)
        inputFile.seek(5)
        producer = MultiPartProducer({
            "field": ('file name', "application/octet-stream", FileBodyProducer(
                      inputFile, cooperator=self.cooperator))})

        # Make sure we are generous enough not to alter seek position:
        self.assertEqual(inputFile.tell(), 5)

        # Total length is hard to calculate manually
        # as it contains a lot of headers parameters, newlines and boundaries
        # let's assert for now that it's no less than the input parameter
        self.assertNotEqual(producer.length, UNKNOWN_LENGTH)
        self.assertTrue(cast(int, producer.length) > len(inputBytes))

        # Calculating length should not touch producers
        self.assertTrue(producer._currentProducer is None)

    def test_defaultCooperator(self) -> None:
        """
        If no L{Cooperator} instance is passed to L{MultiPartProducer}, the
        global cooperator is used.
        """
        producer = MultiPartProducer({
            "field": ("file name", "application/octet-stream", FileBodyProducer(
                      BytesIO(b"yo"),
                      cooperator=self.cooperator))
        })
        self.assertEqual(task.cooperate, producer._cooperate)

    def test_startProducing(self) -> None:
        """
        L{MultiPartProducer.startProducing} starts writing bytes from the input
        file to the given L{IConsumer} and returns a L{Deferred} which fires
        when they have all been written.
        """
        consumer = output = StringTransport()

        # We historically accepted bytes for field names and continue to allow
        # it for compatibility, but the types don't permit it because it makes
        # them even more complicated and awful. So here we verify that that works.
        field = cast(str, b"field")

        producer = MultiPartProducer({
            field: ("file name", "text/hello-world", FileBodyProducer(
                BytesIO(b"Hello, World"),
                cooperator=self.cooperator))
        }, cooperator=self.cooperator, boundary=b"heyDavid")

        complete = producer.startProducing(consumer)

        iterations = 0
        while self._scheduled:
            iterations += 1
            self._scheduled.pop(0)()

        self.assertTrue(iterations > 1)
        self.assertEqual(self.newLines(b"""--heyDavid
Content-Disposition: form-data; name="field"; filename="file name"
Content-Type: text/hello-world
Content-Length: 12

Hello, World
--heyDavid--
"""), output.value())
        self.assertEqual(None, self.successResultOf(complete))

    def test_inputClosedAtEOF(self) -> None:
        """
        When L{MultiPartProducer} reaches end-of-file on the input
        file given to it, the input file is closed.
        """
        inputFile = BytesIO(b"hello, world!")
        consumer = StringTransport()

        producer = MultiPartProducer({
            "field": (
                "file name",
                "text/hello-world",
                FileBodyProducer(
                    inputFile,
                    cooperator=self.cooperator))
        }, cooperator=self.cooperator, boundary=b"heyDavid")

        producer.startProducing(consumer)

        while self._scheduled:
            self._scheduled.pop(0)()

        self.assertTrue(inputFile.closed)

    def test_failedReadWhileProducing(self) -> None:
        """
        If a read from the input file fails while producing bytes to the
        consumer, the L{Deferred} returned by
        L{MultiPartProducer.startProducing} fires with a L{Failure} wrapping
        that exception.
        """
        class BrokenFile:
            def read(self, count):
                raise IOError("Simulated bad thing")

        producer = MultiPartProducer({
            "field": (
                "file name",
                "text/hello-world",
                FileBodyProducer(
                    BrokenFile(),
                    cooperator=self.cooperator))
        }, cooperator=self.cooperator, boundary=b"heyDavid")

        complete = producer.startProducing(StringTransport())

        while self._scheduled:
            self._scheduled.pop(0)()

        self.failureResultOf(complete).trap(IOError)

    def test_stopProducing(self):
        """
        L{MultiPartProducer.stopProducing} stops the underlying
        L{IPullProducer} and the cooperative task responsible for
        calling C{resumeProducing} and closes the input file but does
        not cause the L{Deferred} returned by C{startProducing} to fire.
        """
        inputFile = BytesIO(b"hello, world!")
        consumer = BytesIO()

        producer = MultiPartProducer({
            "field": (
                "file name",
                "text/hello-world",
                FileBodyProducer(
                    inputFile,
                    cooperator=self.cooperator))
        }, cooperator=self.cooperator, boundary=b"heyDavid")
        complete = producer.startProducing(consumer)
        self._scheduled.pop(0)()
        producer.stopProducing()
        self.assertTrue(inputFile.closed)
        self._scheduled.pop(0)()
        self.assertNoResult(complete)

    def test_pauseProducing(self) -> None:
        """
        L{MultiPartProducer.pauseProducing} temporarily suspends writing bytes
        from the input file to the given L{IConsumer}.
        """
        inputFile = BytesIO(b"hello, world!")
        consumer = output = StringTransport()

        producer = MultiPartProducer({
            "field": (
                "file name",
                "text/hello-world",
                FileBodyProducer(
                    inputFile,
                    cooperator=self.cooperator))
        }, cooperator=self.cooperator, boundary=b"heyDavid")
        complete = producer.startProducing(consumer)
        self._scheduled.pop(0)()

        currentValue = output.value()
        self.assertTrue(currentValue)
        producer.pauseProducing()

        # Sort of depends on an implementation detail of Cooperator: even
        # though the only task is paused, there's still a scheduled call.  If
        # this were to go away because Cooperator became smart enough to cancel
        # this call in this case, that would be fine.
        self._scheduled.pop(0)()

        # Since the producer is paused, no new data should be here.
        self.assertEqual(output.value(), currentValue)
        self.assertNoResult(complete)

    def test_resumeProducing(self) -> None:
        """
        L{MultoPartProducer.resumeProducing} re-commences writing bytes
        from the input file to the given L{IConsumer} after it was previously
        paused with L{MultiPartProducer.pauseProducing}.
        """
        inputFile = BytesIO(b"hello, world!")
        consumer = output = StringTransport()

        producer = MultiPartProducer({
            "field": (
                "file name",
                "text/hello-world",
                FileBodyProducer(
                    inputFile,
                    cooperator=self.cooperator))
        }, cooperator=self.cooperator, boundary=b"heyDavid")

        producer.startProducing(consumer)
        self._scheduled.pop(0)()
        currentValue = output.value()
        self.assertTrue(currentValue)
        producer.pauseProducing()
        producer.resumeProducing()
        self._scheduled.pop(0)()
        # make sure we started producing new data after resume
        self.assertTrue(len(currentValue) < len(output.value()))

    def test_unicodeString(self) -> None:
        """
        Make sure unicode string is passed properly
        """
        output, producer = self.getOutput(
            MultiPartProducer({
                "afield": u"Это моя строчечка\r\n",
            }, cooperator=self.cooperator, boundary=b"heyDavid"),
            with_producer=True)

        expected = self.newLines(u"""--heyDavid
Content-Disposition: form-data; name="afield"

Это моя строчечка

--heyDavid--
""".encode("utf-8"))
        self.assertEqual(producer.length, len(expected))
        self.assertEqual(expected, output)

    def test_bytesPassThrough(self) -> None:
        """
        If byte string is passed as a param it is passed through
        unchanged.
        """
        output, producer = self.getOutput(
            MultiPartProducer({
                "bfield": b'\x00\x01\x02\x03',
            }, cooperator=self.cooperator, boundary=b"heyDavid"),
            with_producer=True)

        expected = (
            b"--heyDavid\r\n"
            b'Content-Disposition: form-data; name="bfield"\r\n'
            b'\r\n'
            b'\x00\x01\x02\x03\r\n'
            b'--heyDavid--\r\n'
        )
        self.assertEqual(producer.length, len(expected))
        self.assertEqual(expected, output)

    def test_failOnUnknownParams(self) -> None:
        """
        If byte string is passed as a param and we don't know
        the encoding, fail early to prevent corrupted form posts
        """
        # unknown key
        self.assertRaises(
            ValueError,
            MultiPartProducer, {
                (1, 2): BytesIO(b"yo"),
            },
            cooperator=self.cooperator, boundary=b"heyDavid")

        # tuple length
        self.assertRaises(
            ValueError,
            MultiPartProducer, {
                "a": (1,),
            },
            cooperator=self.cooperator, boundary=b"heyDavid")

        # unknown value type
        self.assertRaises(
            ValueError,
            MultiPartProducer, {
                "a": {"a": "b"},
            },
            cooperator=self.cooperator, boundary=b"heyDavid")

    def test_twoFields(self) -> None:
        """
        Make sure multiple fields are rendered properly.
        """
        output = self.getOutput(
            MultiPartProducer({
                "afield": "just a string\r\n",
                "bfield": "another string"
            }, cooperator=self.cooperator, boundary=b"heyDavid"))

        self.assertEqual(self.newLines(b"""--heyDavid
Content-Disposition: form-data; name="afield"

just a string

--heyDavid
Content-Disposition: form-data; name="bfield"

another string
--heyDavid--
"""), output)

    def test_fieldsAndAttachment(self):
        """
        Make sure multiple fields are rendered properly.
        """
        output, producer = self.getOutput(
            MultiPartProducer({
                "bfield": "just a string\r\n",
                "cfield": "another string",
                "afield": (
                    "file name",
                    "text/hello-world",
                    FileBodyProducer(
                        inputFile=BytesIO(b"my lovely bytes"),
                        cooperator=self.cooperator))
            }, cooperator=self.cooperator, boundary=b"heyDavid"),
            with_producer=True)

        expected = self.newLines(b"""--heyDavid
Content-Disposition: form-data; name="bfield"

just a string

--heyDavid
Content-Disposition: form-data; name="cfield"

another string
--heyDavid
Content-Disposition: form-data; name="afield"; filename="file name"
Content-Type: text/hello-world
Content-Length: 15

my lovely bytes
--heyDavid--
""")
        self.assertEqual(producer.length, len(expected))
        self.assertEqual(output, expected)

    def test_multipleFieldsAndAttachments(self):
        """
        Make sure multiple fields, attachments etc are rendered properly.
        """
        output, producer = self.getOutput(
            MultiPartProducer({
                "cfield": "just a string\r\n",
                "bfield": "another string",
                "efield": (
                    "ef",
                    "text/html",
                    FileBodyProducer(
                        inputFile=BytesIO(b"my lovely bytes2"),
                        cooperator=self.cooperator)),
                "xfield": (
                    "xf",
                    "text/json",
                    FileBodyProducer(
                        inputFile=BytesIO(b"my lovely bytes219"),
                        cooperator=self.cooperator)),
                "afield": (
                    "af",
                    "text/xml",
                    FileBodyProducer(
                        inputFile=BytesIO(b"my lovely bytes22"),
                        cooperator=self.cooperator))
            }, cooperator=self.cooperator, boundary=b"heyDavid"),
            with_producer=True)

        expected = self.newLines(b"""--heyDavid
Content-Disposition: form-data; name="bfield"

another string
--heyDavid
Content-Disposition: form-data; name="cfield"

just a string

--heyDavid
Content-Disposition: form-data; name="afield"; filename="af"
Content-Type: text/xml
Content-Length: 17

my lovely bytes22
--heyDavid
Content-Disposition: form-data; name="efield"; filename="ef"
Content-Type: text/html
Content-Length: 16

my lovely bytes2
--heyDavid
Content-Disposition: form-data; name="xfield"; filename="xf"
Content-Type: text/json
Content-Length: 18

my lovely bytes219
--heyDavid--
""")
        self.assertEqual(producer.length, len(expected))
        self.assertEqual(output, expected)

    def test_unicodeAttachmentName(self):
        """
        Make sure unicode attachment names are supported.
        """
        output, producer = self.getOutput(
            MultiPartProducer({
                "field": (
                    u'Так себе имя.jpg',
                    "image/jpeg",
                    FileBodyProducer(
                        inputFile=BytesIO(b"my lovely bytes"),
                        cooperator=self.cooperator
                    )
                )
            }, cooperator=self.cooperator, boundary=b"heyDavid"),
            with_producer=True)

        expected = self.newLines(u"""--heyDavid
Content-Disposition: form-data; name="field"; filename="Так себе имя.jpg"
Content-Type: image/jpeg
Content-Length: 15

my lovely bytes
--heyDavid--
""".encode("utf-8"))
        self.assertEqual(len(expected), producer.length)
        self.assertEqual(expected, output)

    def test_missingAttachmentName(self):
        """
        Make sure attachments without names are supported
        """
        output, producer = self.getOutput(
            MultiPartProducer({
                "field": (
                    None,
                    "image/jpeg",
                    FileBodyProducer(
                        inputFile=BytesIO(b"my lovely bytes"),
                        cooperator=self.cooperator,
                    )
                )
            }, cooperator=self.cooperator,
                boundary=b"heyDavid"),
            with_producer=True)

        expected = self.newLines(b"""--heyDavid
Content-Disposition: form-data; name="field"
Content-Type: image/jpeg
Content-Length: 15

my lovely bytes
--heyDavid--
""")
        self.assertEqual(len(expected), producer.length)
        self.assertEqual(expected, output)

    def test_newLinesInParams(self):
        """
        Make sure we generate proper format even with newlines in attachments
        """
        output = self.getOutput(
            MultiPartProducer({
                "field": (
                    u'\r\noops.j\npg',
                    "image/jp\reg\n",
                    FileBodyProducer(
                        inputFile=BytesIO(b"my lovely bytes"),
                        cooperator=self.cooperator
                    )
                )
            }, cooperator=self.cooperator,
                boundary=b"heyDavid"
            )
        )

        self.assertEqual(self.newLines(u"""--heyDavid
Content-Disposition: form-data; name="field"; filename="oops.jpg"
Content-Type: image/jpeg
Content-Length: 15

my lovely bytes
--heyDavid--
""".encode("utf-8")), output)

    def test_worksWithMultipart(self):
        """
        Make sure the stuff we generated can actually be parsed by the
        `multipart` module.
        """
        output = self.getOutput(
            MultiPartProducer([
                ("cfield", "just a string\r\n"),
                ("cfield", "another string"),
                ("efield", ('ef', "text/html", FileBodyProducer(
                            inputFile=BytesIO(b"my lovely bytes2"),
                            cooperator=self.cooperator,
                            ))),
                ("xfield", ('xf', "text/json", FileBodyProducer(
                            inputFile=BytesIO(b"my lovely bytes219"),
                            cooperator=self.cooperator,
                            ))),
                ("afield", ('af', "text/xml", FileBodyProducer(
                            inputFile=BytesIO(b"my lovely bytes22"),
                            cooperator=self.cooperator,
                            )))
            ], cooperator=self.cooperator, boundary=b"heyDavid"
            )
        )

        form = MultipartParser(
            stream=BytesIO(output),
            boundary=b"heyDavid",
            content_length=len(output),
        )

        self.assertEqual(
            [b'just a string\r\n', b'another string'],
            [f.raw for f in form.get_all('cfield')],
        )

        self.assertEqual(b'my lovely bytes2', form.get('efield').raw)
        self.assertEqual(b'my lovely bytes219', form.get('xfield').raw)
        self.assertEqual(b'my lovely bytes22', form.get('afield').raw)


class LengthConsumerTestCase(unittest.TestCase):
    """
    Tests for the _LengthConsumer, an L{IConsumer} which is used to compute
    the length of a produced content.
    """

    def test_scalarsUpdateCounter(self):
        """
        When an int is written, _LengthConsumer updates its internal
        counter.
        """
        consumer = _LengthConsumer()
        self.assertEqual(consumer.length, 0)
        consumer.write(1)
        self.assertEqual(consumer.length, 1)
        consumer.write(2147483647)
        self.assertEqual(consumer.length, 2147483648)

    def test_stringUpdatesCounter(self):
        """
        Use the written string length to update the internal counter
        """
        a = (b"Cantami, o Diva, del Pelide Achille\n l'ira funesta che "
             b"infiniti addusse\n lutti agli Achei")

        consumer = _LengthConsumer()
        self.assertEqual(consumer.length, 0)
        consumer.write(a)
        self.assertEqual(consumer.length, 89)



"""
Utilities related to retrieving the contents of the response-body.
"""

import json
from typing import Any, Callable, FrozenSet, List, Optional, cast

from ._multipart import parse_options_header
from twisted.internet.defer import Deferred, succeed
from twisted.internet.protocol import Protocol, connectionDone
from twisted.python.failure import Failure
from twisted.web.client import ResponseDone
from twisted.web.http import PotentialDataLoss
from twisted.web.http_headers import Headers
from twisted.web.iweb import IResponse


"""Characters that are valid in a charset name per RFC 2978.

See https://www.rfc-editor.org/errata/eid5433
"""
_MIME_CHARSET_CHARS: FrozenSet[str] = frozenset(
    "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"  # ALPHA
    "0123456789"  # DIGIT
    "!#$%&+-^_`~"  # symbols
)


def _encoding_from_headers(headers: Headers) -> Optional[str]:
    content_types = headers.getRawHeaders("content-type")
    if content_types is None:
        return None

    # This seems to be the choice browsers make when encountering multiple
    # content-type headers.
    media_type, params = parse_options_header(content_types[-1])

    charset = params.get("charset")
    if charset:
        assert isinstance(charset, str)  # for MyPy
        charset = charset.strip("'\"").lower()
        if not charset:
            return None
        if not set(charset).issubset(_MIME_CHARSET_CHARS):
            return None
        return charset

    if media_type == "application/json":
        return "utf-8"

    return None


class _BodyCollector(Protocol):
    finished: "Optional[Deferred[None]]"

    def __init__(
        self, finished: "Deferred[None]", collector: Callable[[bytes], None]
    ) -> None:
        self.finished = finished
        self.collector = collector

    def dataReceived(self, data: bytes) -> None:
        try:
            self.collector(data)
        except BaseException:
            if self.transport:
                self.transport.loseConnection()
            if self.finished:
                self.finished.errback(Failure())
            self.finished = None

    def connectionLost(self, reason: Failure = connectionDone) -> None:
        if self.finished is None:
            return
        if reason.check(ResponseDone):
            self.finished.callback(None)
        elif reason.check(PotentialDataLoss):
            # http://twistedmatrix.com/trac/ticket/4840
            self.finished.callback(None)
        else:
            self.finished.errback(reason)


def collect(
    response: IResponse, collector: Callable[[bytes], None]
) -> "Deferred[None]":
    """
    Incrementally collect the body of the response.

    This function may only be called **once** for a given response.

    If the ``collector`` raises an exception, it will be set as the error value
    on response ``Deferred`` returned from this function, and the underlying
    HTTP transport will be closed.

    :param IResponse response: The HTTP response to collect the body from.
    :param collector: A callable to be called each time data is available from
        the response body.
    :type collector: single argument callable

    :rtype: Deferred that fires with None when the entire body has been read.
    """
    if response.length == 0:
        return succeed(None)

    d: "Deferred[None]" = Deferred()
    response.deliverBody(_BodyCollector(d, collector))
    return d


def content(response: IResponse) -> "Deferred[bytes]":
    """
    Read the contents of an HTTP response.

    This function may be called multiple times for a response, it uses a
    ``WeakKeyDictionary`` to cache the contents of the response.

    :param IResponse response: The HTTP Response to get the contents of.

    :rtype: Deferred that fires with the content as a str.
    """
    _content: List[bytes] = []
    d = collect(response, _content.append)
    return cast(
        "Deferred[bytes]",
        d.addCallback(lambda _: b"".join(_content)),
    )


def json_content(response: IResponse, **kwargs: Any) -> "Deferred[Any]":
    """
    Read the contents of an HTTP response and attempt to decode it as JSON.

    This function relies on :py:func:`content` and so may be called more than
    once for a given response.

    :param IResponse response: The HTTP Response to get the contents of.

    :param kwargs: Any keyword arguments accepted by :py:func:`json.loads`

    :rtype: Deferred that fires with the decoded JSON.
    """
    # RFC7159 (8.1): Default JSON character encoding is UTF-8
    d = text_content(response, encoding="utf-8")
    return d.addCallback(lambda text: json.loads(text, **kwargs))


def text_content(response: IResponse, encoding: str = "ISO-8859-1") -> "Deferred[str]":
    """
    Read the contents of an HTTP response and decode it with an appropriate
    charset, which may be guessed from the ``Content-Type`` header.

    :param IResponse response: The HTTP Response to get the contents of.
    :param str encoding: A charset, such as ``UTF-8`` or ``ISO-8859-1``,
        used if the response does not specify an encoding.

    :rtype: Deferred that fires with a unicode string.
    """

    def _decode_content(c: bytes) -> str:

        e = _encoding_from_headers(response.headers)

        if e is not None:
            return c.decode(e)

        return c.decode(encoding)

    d = content(response)
    return cast("Deferred[str]", d.addCallback(_decode_content))



"""
Convenience helpers for :mod:`http.cookiejar`
"""

from typing import Union, Iterable, Optional
from http.cookiejar import Cookie, CookieJar

from hyperlink import EncodedURL


def scoped_cookie(origin: Union[str, EncodedURL], name: str, value: str) -> Cookie:
    """
    Create a cookie scoped to a given URL's origin.

    You can insert the result directly into a `CookieJar`, like::

        jar = CookieJar()
        jar.set_cookie(scoped_cookie("https://example.tld", "flavor", "chocolate"))

        await treq.get("https://domain.example", cookies=jar)

    :param origin:
        A URL that specifies the domain and port number of the cookie.

        If the protocol is HTTP*S* the cookie is marked ``Secure``, meaning
        it will not be attached to HTTP requests. Otherwise the cookie will be
        attached to both HTTP and HTTPS requests

    :param name: Name of the cookie.

    :param value: Value of the cookie.

    .. note::

        This does not scope the cookies to any particular path, only the
        host, port, and scheme of the given URL.
    """
    if isinstance(origin, EncodedURL):
        url_object = origin
    else:
        url_object = EncodedURL.from_text(origin)

    secure = url_object.scheme == "https"
    port_specified = not (
        (url_object.scheme == "https" and url_object.port == 443)
        or (url_object.scheme == "http" and url_object.port == 80)
    )
    port = str(url_object.port) if port_specified else None
    domain = url_object.host
    netscape_domain = domain if "." in domain else domain + ".local"
    return Cookie(
        # Scoping
        domain=netscape_domain,
        port=port,
        secure=secure,
        port_specified=port_specified,
        # Contents
        name=name,
        value=value,
        # Constant/always-the-same stuff
        version=0,
        path="/",
        expires=None,
        discard=False,
        comment=None,
        comment_url=None,
        rfc2109=False,
        path_specified=False,
        domain_specified=False,
        domain_initial_dot=False,
        rest={},
    )


def search(
    jar: CookieJar, *, domain: str, name: Optional[str] = None
) -> Iterable[Cookie]:
    """
    Raid the cookie jar for matching cookies.

    This is O(n) on the number of cookies in the jar.

    :param jar: The `CookieJar` (or subclass thereof) to search.

    :param domain:
        Domain, as in the URL, to match. ``.local`` is appended to
        a bare hostname. Subdomains are not matched (i.e., searching
        for ``foo.bar.tld`` won't return a cookie set for ``bar.tld``).

    :param name: Cookie name to match (exactly)
    """
    netscape_domain = domain if "." in domain else domain + ".local"

    for c in jar:
        if c.domain != netscape_domain:
            continue
        if name is not None and c.name != name:
            continue
        yield c



# Copyright (c) The treq Authors.
# See LICENSE for details.
import io
from http.cookiejar import CookieJar
from typing import Any, Dict, Iterable, List, Mapping, Tuple, Union

from hyperlink import DecodedURL, EncodedURL
from twisted.internet.interfaces import (IReactorPluggableNameResolver,
                                         IReactorTCP, IReactorTime)
from twisted.web.http_headers import Headers
from twisted.web.iweb import IBodyProducer


class _ITreqReactor(IReactorTCP, IReactorTime, IReactorPluggableNameResolver):
    """
    The kind of reactor treq needs for type-checking purposes.

    This is an approximation of the actual requirement, which comes from the
    `twisted.internet.endpoints.HostnameEndpoint` used by the `Agent`
    implementation:

    > Provider of IReactorTCP, IReactorTime and either
    > IReactorPluggableNameResolver or IReactorPluggableResolver.

    We don't model the `IReactorPluggableResolver` option because it is
    deprecated.
    """


_S = Union[bytes, str]

_URLType = Union[
    str,
    bytes,
    EncodedURL,
    DecodedURL,
]

_ParamsType = Union[
    Mapping[str, Union[str, Tuple[str, ...], List[str]]],
    List[Tuple[str, str]],
]

_HeadersType = Union[
    Headers,
    Dict[_S, _S],
    Dict[_S, List[_S]],
]

_CookiesType = Union[
    CookieJar,
    Mapping[str, str],
]

_WholeBody = Union[
    bytes,
    io.BytesIO,
    io.BufferedReader,
    IBodyProducer,
]
"""
Types that define the entire HTTP request body, including those coercible to
`IBodyProducer`.
"""

# Concrete types are used here because the handling of the *data* parameter
# does lots of isinstance checks.
_BodyFields = Union[
    Dict[str, str],
    List[Tuple[str, str]],
]
"""
Types that will be URL- or multipart-encoded before being sent as part of the
HTTP request body.
"""

_DataType = Union[_WholeBody, _BodyFields]
"""
Values accepted for the *data* parameter

Note that this is a simplification. Only `_BodyFields` may be supplied if the
*files* parameter is passed.
"""

_FileValue = Union[
    str,
    bytes,
    Tuple[str, str, IBodyProducer],
]
"""
Either a scalar string, or a file to upload as (filename, content type,
IBodyProducer)
"""

_FilesType = Union[
    Mapping[str, _FileValue],
    Iterable[Tuple[str, _FileValue]],
]
"""
Values accepted for the *files* parameter.
"""

# Soon... 🤞 https://github.com/python/mypy/issues/731
_JSONType = Any
